import os
import numbers

from . import file_ops as fo
from . import image as ims


def parse_arguments():
    import argparse

    parser_x = argparse.ArgumentParser()
    parser_y = parser_x.add_mutually_exclusive_group()
    lazy_parser = parser_y.add_argument_group()
    lazy_choices = lazy_parser.add_mutually_exclusive_group()

    lazy_choices.add_argument(
        "--vstack",
        nargs="+",
        type=str,
        help="List any images you want to stack vertically",
    )

    lazy_choices.add_argument(
        "--hstack",
        nargs="+",
        type=str,
        help="List any images you want to stack horizontally",
    )

    parser = parser_y.add_argument_group()
    # TODO: excess images in directory? Insufficient images in directory?

    sources = parser.add_mutually_exclusive_group()

    sources.add_argument(
        "--dirs_in",
        dest="dirs_in",
        nargs="+",
        type=str,
        help="List any directories you want to use.",
    )

    sources.add_argument(
        "--files_in",
        dest="files_in",
        nargs="+",
        type=str,
        help="List any images or videos you want to use. Not compatible"
        "with --dirs_in.",
    )

    parser.add_argument(
        "--ext_in",
        dest="ext_in",
        type=str,
        nargs="*",
        choices=["png", "jpg", "mp4"],
        default=["png", "jpg"],
        help="Will select files with given extension for input. Ignored if --files_in provided.",
    )

    parser.add_argument(
        "--ext_out",
        dest="ext_out",
        type=str,
        choices=["png", "jpg", "mp4"],
        default=None,
        help="Outputs file with given extension."
        " Overridden by --to_vid, --to_img, --to_imgs, and set from --name by default.",
    )

    parser.add_argument(
        "--dir_out",
        dest="dir_out",
        type=str,
        default=None,
        help="Output directory. Set from --name by default.",
    )

    parser.add_argument(
        "--name",
        dest="name",
        type=str,
        default="./output/file.jpg",
        help="File name. Sets --dir_out and --ext_out if not already specified.",
    )

    output_formats = parser.add_mutually_exclusive_group()

    output_formats.add_argument(
        "--to_vid",
        dest="to_vid",
        action="store_true",
        help="Will output a video file (default 30fps, .mp4). Not compatible with --to_img."
        " If multiple directories are provided, will only use first x*y videos."
        " If one directory is provided, will use first x*y videos.",
    )

    output_formats.add_argument(
        "--to_imgs",
        dest="to_imgs",
        action="store_true",
        help="Will output many image files (default .jpg)",
    )

    resize = parser.add_mutually_exclusive_group()
    resize.add_argument(
        "--resize_in",
        dest="resize_in",
        nargs=2,
        type=int,
        help="Sets the dimensions of the input image. "
        "Not compatible with -resize_out or -resize_down or -resize_up",
    )

    resize.add_argument(
        "--resize_out",
        dest="resize_out",
        nargs=2,
        type=int,
        help="Sets the dimensions of the output image. "
        "Not compatible with -resize_in or -resize_down or -resize_up",
    )

    resize.add_argument(
        "--resize",
        dest="resize",
        choices=[ims.Resize.UP, ims.Resize.DOWN, ims.Resize.FIRST],
        type=get_resize_enum,
        default="first",
        help="Resizes the images according to the choice. "
        "down resizes each image to the smallest image. "
        "up resizes each image to the largest image. "
        "first resizes each image to the first image seen.",
    )

    parser.add_argument(
        "--cols",
        dest="cols",
        type=int,
        default=1,
        help="Number of images or videos placed side by side, horizontally.",
    )

    parser.add_argument(
        "--rows",
        dest="rows",
        type=int,
        default=1,
        help="Number of images or videos stacked on top of each other," " vertically.",
    )

    parser.add_argument(
        "--fps",
        dest="fps",
        default=30.0,
        type=float,
        help="Frame rate of output video. Not compatible if videos are passed in.",
    )

    parser.add_argument(
        "--fps_unlock",
        dest="fps_unlock",
        action="store_true",
        help="Removes the requirement for input videos to have matching framerates. "
        "May cause issues with frames not matching up if framerates do not match.",
    )

    parser.add_argument(
        "--max",
        dest="max_imgs",
        default=None,
        type=int,
        help="Maximum number of images to output (eg. if folder has 1000 images, output only 10.",
    )

    parser.add_argument(
        "--read_matching_file_names",
        action="store_true",
        help="Will concatenate files with the same name from each directory,"
        " and will resize on a per image basis unless a width and height are specified.",
    )

    return validate_arguments(parser_x)


def mixed_ext(exts) -> bool:
    return fo.has_image_exts(exts) and fo.has_video_exts(exts)


# Assumes args has resize_up, resize_down, and resize_first
def get_resize_enum(s) -> ims.Resize:
    """
    Identifies which option has been selected.

    :param s:
    :return:
    """
    return {
        "up": ims.Resize.UP,
        "down": ims.Resize.DOWN,
        "first": ims.Resize.FIRST,
    }.get(s, ims.Resize.NONE)


class IsPlural:
    def __init__(self, pl):
        if hasattr(pl, "__len__") and (not isinstance(pl, str)):
            self.is_plural = len(pl) != 1
        elif isinstance(pl, numbers.Number):
            self.is_plural = abs(pl) != 1
        else:
            self.is_plural = False


class PluralizableString:
    def __init__(
        self,
        base_string: str,
        non_plural_end: str,
        plural_end: str,
        pluralizable: IsPlural,
    ):
        self.text: str = base_string + (
            plural_end if pluralizable.is_plural else non_plural_end
        )

    def __repr__(self):
        return self.text


def parse_lazy_arguments(args):
    """
    Checks that the arguments for lazy image manipulation are correct.

    :param args:
    :return:        A boolean indicating whether lazy arguments exist.
    """
    if not (args.vstack or args.hstack):
        return

    args.files_in = args.vstack or args.hstack
    args.vstack = bool(args.vstack)
    args.hstack = bool(args.hstack)
    args.cols, args.rows = (
        (len(args.files_in), 1) if args.hstack else (1, len(args.files_in))
    )


def validate_dirs(parser, args):
    if not args.dirs_in:
        return

    if args.to_imgs and not args.ext_in:
        parser.error("No extension specified for input images.")

    missing_dirs = fo.get_missing_dirs(args.dirs_in)

    if missing_dirs:
        pl = IsPlural(missing_dirs)
        dir_str = PluralizableString("director", "y", "ies", pl)
        do_str = PluralizableString("do", "es", "", pl)
        parser.error(
            "The %s specified at %s %s not exist."
            % (dir_str, ", ".join(missing_dirs), do_str)
        )


def validate_files(parser, args):
    if not args.files_in:
        return

    image_delta = len(args.files_in) - args.cols * args.rows

    if image_delta < 0:
        parser.error(
            "Not enough photos given to generate an image or video stacked %d by %d (requires %d images or "
            "videos, %d given)"
            % (args.cols, args.rows, args.cols * args.rows, len(args.files_in))
        )

    if image_delta > 0:
        print("Last %d files will not be included in output" % image_delta)

    missing_files = fo.get_missing_files(args.files_in)

    if missing_files:
        pl = IsPlural(missing_files)
        dir_str = PluralizableString("file", "", "s", pl)
        do_str = PluralizableString("do", "es", "", pl)
        parser.error(
            "The %s specified at %s %s not exist."
            % (dir_str, ", ".join(missing_files), do_str)
        )


def validate_extensions(parser, args):
    """
    Validate that the extensions provided are correct and are not mixed.

    :param parser:
    :param args:
    :return:
    """
    if args.files_in:
        args.ext_in = {fo.get_ext(file) for file in args.files_in}

    if mixed_ext(args.ext_in):
        parser.error("Must have exclusively images or videos as input.")

    if not args.dirs_in:
        return

    if args.to_vid and args.ext_out not in {"mp4"}:
        args.ext_out = "mp4"
        print("Output extension automatically set to %s." % args.ext_out)


def validate_arguments(parser):
    """
    Checks that all file paths are correct and that no conflicting variables exist.

    :param parser:  An argparse.ArgumentParser object.
    :return:        A set of parsed and valid arguments.
    """

    args = parser.parse_args()
    args.dir_out = args.dir_out or os.path.dirname(args.name) or "./"
    args.ext_out = args.ext_out or fo.get_ext(args.name)
    args.name = ".".join(os.path.splitext(os.path.basename(args.name))[:-1])

    parse_lazy_arguments(args)

    validate_extensions(parser, args)
    validate_dirs(parser, args)
    validate_files(parser, args)

    if args.read_matching_file_names:
        if not args.dirs_in:
            parser.error(
                "--read_matching_file_names requires --dirs_in to be specified."
            )
        if args.cols * args.rows != len(args.dirs_in):
            parser.error(
                f"Provided --dirs_in of {len(args.dirs_in)} directories and output stacking"
                f" of {args.cols} by {args.rows} images are not compatible."
            )

    return args



from typing import Union, List, Tuple

import cv2

from . import file_ops as fo
from . import image as ims
from .image import Stacking


def video_dimensions(video) -> Tuple[int, int]:
    return int(video.get(cv2.CAP_PROP_FRAME_WIDTH)), int(
        video.get(cv2.CAP_PROP_FRAME_HEIGHT)
    )


class VideoIterator(ims.GenericImageIterator):
    def __init__(
        self,
        paths: Union[List[str], str],
        stacking: Stacking = None,
        lock_framerate: bool = True,
    ):
        """
        Initializes a video iterator that will return num frames per iteration from each video in paths_to_videos.

        :param paths:     The file paths to the videos to read
        :param num:                 The number of frames to return each iteration.
        """
        if isinstance(paths, str):
            paths = [paths]

        super().__init__(stacking=stacking or Stacking(len(paths), 1, "rd"))

        self.paths = paths

        if not fo.check_files_exist(paths):
            raise ValueError("One or more videos not found.")

        self.videos = []
        self.frame_index = 0
        self.completed_videos = set()
        self.last_frame = [None] * len(self.paths)
        self.fps = None
        self.frame_lock = lock_framerate
        self._load_videos()

    def _set_dims(self):
        """
        If one or more dimensions are not already set, sets the image dimensions automatically.
        Assumes that all videos have been initialized.
        """
        self.resize_in(
            ims.Resize.DOWN.choose(video_dimensions(video) for video in self.videos)
        )

    def _load_video(self, counter):
        video = cv2.VideoCapture(self.paths[counter])
        self.videos.append(video)
        if self.frame_lock:
            if self.fps not in {None, video.get(cv2.CAP_PROP_FPS)}:
                raise ValueError("Video FPS does not match.")
            else:
                self.fps = video.get(cv2.CAP_PROP_FPS)

    def _load_videos(self):
        """
        Loads the videos into memory, initializes variables.

        :return:
        """
        for counter in range(len(self.paths)):
            self._load_video(counter)
        self._max_iters = int(
            max(video.get(cv2.CAP_PROP_FRAME_COUNT) for video in self.videos)
        )
        self._set_dims()

    def __iter__(self):
        return self

    # Returns num frames from all videos. If one video has reached the end, will keep last frame.
    def __next__(self):
        if self._num_iters >= self._max_iters:
            raise StopIteration()

        output = []
        for i, video in enumerate(self.videos):
            if i not in self.completed_videos:
                success, image = video.read()
                if success:
                    self.last_frame[i] = image
                else:
                    video.release()
                    self.completed_videos.add(i)
            output.append(self.last_frame[i])

        name, ext = self._name_file()
        images = ims.ImageDataStore(output, name, ext)
        self._num_iters += 1
        self.frame_index += 1

        return self._apply_transforms(images)

    def skip(self, count: int):
        self.frame_index += count
        for i, video in enumerate(self.videos):
            if i not in self.completed_videos:
                video.set(cv2.CAP_PROP_POS_FRAMES, self.frame_index)
        return self

    def take(self, count: int):
        self._max_iters = count
        return self



def validate_resize_out(cols, rows, dims):
    width, height = dims

    if width % cols != 0:
        raise EnvironmentError(
            "Output width must be a multiple of image stacking number."
        )

    if height % rows != 0:
        raise EnvironmentError(
            "Output height must be a multiple of image stacking number."
        )

    return width // cols, height // rows


def get_correct_dimensions(args):
    if args.resize_in:
        return args.resize_in

    if args.resize_out:
        return validate_resize_out(args.cols, args.rows, args.resize_out)

    if args.read_matching_file_names:
        return None

    if args.dirs_in:
        return ims.get_dimensions_dirs(args.dirs_in, args.ext_in, args.resize)

    if not args.files_in:
        raise EnvironmentError("No files provided.")

    return args.resize.choose(ims.dimensions(args.files_in))


if __name__ == "__main__":
    from . import arg_parser as ap
    from . import video as vs
    from . import image as ims
    from . import file_ops as fo

    import os

    args = ap.parse_arguments()

    input_args = [args.dirs_in, args.ext_in] if args.dirs_in else [args.files_in]
    size = get_correct_dimensions(args)

    vargs = {
        "stacking": ims.Stacking(args.cols, args.rows, "rd"),
    }

    os.makedirs(os.path.dirname(args.dir_out), exist_ok=True)

    if args.read_matching_file_names:
        source = ims.DirectoryIteratorMatchNames(
            args.dirs_in, **vargs
        ).resize_individual(args.resize)
    elif args.dirs_in:
        if fo.has_image_exts(args.ext_in):
            source = ims.DirectoryIterator(
                args.dirs_in, args.ext_in, **vargs
            ).resize_in(size)
        else:
            files_in = fo.get_first_n_files(
                args.dirs_in, args.ext_in, args.cols * args.rows
            )
            if len(files_in) != args.cols * args.rows:
                raise ValueError(
                    "Insufficient files found in %s" % ", ".join(args.dirs_in)
                )
            print(
                "Automatically selected these video files to concatenate: %s"
                % (", ".join(files_in))
            )
            source = vs.VideoIterator(files_in, **vargs).resize_in(size)
    else:
        if fo.has_image_exts(args.ext_in):
            source = ims.FileIterator(args.files_in, **vargs).resize_in(size)
        else:
            source = vs.VideoIterator(
                args.files_in, lock_framerate=not args.fps_unlock, **vargs
            ).resize_in(size)

    if args.max_imgs:
        source = source.take(args.max_imgs)

    print(
        "Output file will have dimensions: %d x %d px."
        % (size[0] * args.cols, size[1] * args.rows)
    )

    if args.to_imgs:
        source = source.rename(args.dir_out, args.name, args.ext_out)
        source.write_images()
    elif args.to_vid:
        source.write_video(
            fo.form_file_name(args.dir_out, args.name, args.ext_out), fps=args.fps
        )
    else:
        raise NotImplementedError()



import enum
import glob
import os
from pathlib import Path

from typing import Union, List
from os.path import isdir, isfile


def append_forward_slash_path(path: str) -> str:
    """
    Appends a forward slash to the path if one is not present.
    :return:
    """
    return path if path[-1] == "/" else path + "/"


def get_missing_files(files: Union[str, List[str]]) -> List[str]:
    """
    Returns all missing files.

    :param files: one or more files to check.
    :return:
    """
    if isinstance(files, str):
        return [] if isfile(files) else [files]

    return [file for file in files if not isfile(file)]


def check_files_exist(files: Union[str, List[str]]) -> bool:
    """
    Returns true if all files exist, otherwise false.

    :param files: one or more files to check.
    :return:
    """
    if isinstance(files, str):
        return isfile(files)

    return all(isfile(file) for file in files)


def get_missing_dirs(directories: Union[str, List[str]]) -> List[str]:
    """
    Returns all directories from directories which don't exist.

    :param directories: one or more directories to check.
    :return:
    """
    if isinstance(directories, str):
        return [] if isdir(directories) else [directories]

    return [directory for directory in directories if not isdir(directory)]


def check_dirs_exist(directories: Union[str, List[str]]) -> bool:
    """
    Returns true if all directories exist, otherwise false.

    :param      directories: one or more directories to check.
    :return:
    """
    if isinstance(directories, str):
        return isdir(directories)

    return all(isdir(directory) for directory in directories)


def match_all_cases(strx: str):
    """
    Exists to provide compatibility for Unix systems.
    Source: https://stackoverflow.com/questions/8151300/ignore-case-in-glob-on-linux

    :param strx: Any string.
    :return:     Returns a regex which will match the same string.
    """
    return "".join(
        "[%s%s]" % (c.lower(), c.upper()) if c.isalpha() else c for c in strx
    )


def get_files(directory: str, extensions: Union[List[str], str]) -> List[str]:
    """
    Returns a list of file names in the given directory ending in the given extensions

    :param directory:   one or more directories to search.
    :param extensions:  one or more extensions to match.
    :return:
    """

    if isinstance(extensions, str):
        jpg_subset = [".jpeg", ".jpg"]
        extensions = {
            "jpg": jpg_subset,
            "jpeg": jpg_subset,
            "png": [".png"],
            "mp4": [".mp4"],
        }.get(extensions.lower().lstrip("."), [extensions])

    directory = append_forward_slash_path(directory)
    extensions = [prepend_dot(ext) for ext in extensions]
    frames = {
        file
        for ext in extensions
        # TODO: when Python min_ver is 3.10, use root_dir=directory
        for file in glob.glob(f"{directory}*{match_all_cases(ext)}")
    }

    return sorted(list(frames))


def get_first_n_files(
    directories: Union[List[str], str], ext: Union[List[str], str], num: int
) -> List[str]:
    """
    Returns the first n files in the directories that match the given extensions, as evenly as possible.
    In the event that less than num files exist, will return all matches found.

    :param directories: one or more directories to search.
    :param ext: one or more extensions to match.
    :param num: number of files that should be matched and returned.
    :return:
    """
    if not isinstance(directories, list):
        directories = [directories]
    dirs = [get_files(directory, ext) for directory in directories]

    exhausted_dirs = set()

    curr_dir = 0
    curr_index = 0
    output = []
    while len(output) < num:
        # Directory has no images left
        if curr_index >= len(dirs[curr_dir]):
            exhausted_dirs.add(curr_dir)
            # Given directory is "exhausted": no images left
            exhausted_dirs.add(curr_dir)
            # All directories are exhausted.
            if len(exhausted_dirs) == len(dirs):
                return output
        else:
            # Otherwise, append next image in directory.
            output.append(dirs[curr_dir][curr_index])
        # Move to next image in directory.
        curr_dir += 1
        if curr_dir % len(dirs) == 0:
            curr_index += 1
            curr_dir = 0
    return output


def prepend_dot(ext: str) -> str:
    return ext if ext[0] == "." else "." + ext


def clear_files(directory: str, *exts) -> None:
    """
    Clears the given folder of any and all files that match any extension provided.
    :param directory: folder to remove extensions from.
    :param exts: one or more extensions.
    :return:
    """
    directory = append_forward_slash_path(directory)
    for ext in exts:
        ext = prepend_dot(ext)
        # TODO: when Python min_ver is 3.10, use root_dir=directory
        for file in glob.glob(f"{directory}*{ext}"):
            os.remove(file)


def form_file_name(dir_out: str, file_name: str, ext: str) -> str:
    """
    Removes excess extensions in the file_name and returns a fully formed file name, cleaned of excess extensions.
    :param dir_out:     path to a directory.
    :param file_name:   A file name with zero or more extensions.
    :param ext:         the file extension
    :return:
    """
    return str((Path(dir_out) / file_name).with_suffix(prepend_dot(ext)))


def get_ext(file: str):
    """
    Returns the file extension without any preceding dots.
    :param file:    The file from which to get the extension.
    :return:
    """
    return Path(file).suffix.lstrip(".")


class FileCategory(enum.Enum):
    VIDEO = 0
    IMAGE = 1

    @classmethod
    def from_str(cls, s: str) -> "FileCategory":
        return {"mp4": cls.VIDEO, "png": cls.IMAGE, "jpg": cls.IMAGE}[s]


def has_video_exts(exts: Union[List[str], str]):
    video_exts = {"mp4"}
    if isinstance(exts, str):
        return exts in video_exts
    return bool(set(exts).intersection(video_exts))


def has_image_exts(exts: Union[List[str], str]):
    image_exts = {"png", "jpg"}
    if isinstance(exts, str):
        return exts in image_exts
    return bool(set(exts).intersection(image_exts))



import setuptools
from setuptools import find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

print(find_packages())

setuptools.setup(
    name="imgvid_utils",
    version="0.2.1",
    author="Philippe Solodov",
    author_email="solop1906@gmail.com",
    description="A package that provides helpful utilities to interact with videos and images through OpenCV.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/philippeitis/imgvid_utils",
    packages=find_packages(exclude=["tests", "tests.*"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        "opencv-python==4.6.0.66",
    ],

)



from pathlib import Path

import cv2
from enum import Enum
from typing import Union, List, Tuple, Iterable, Optional
import os
import numpy as np

from . import file_ops as fo


def find(items: List, predicate) -> Optional[int]:
    """
    Returns the first index of an item in items that satisfies predicate
    :param items:
    :param predicate:
    :return:
    """
    for index, item in enumerate(items):
        if predicate(item):
            return index
    return None


class SkipIterator:
    def __init__(self, items: List, skip: int = 0):
        self.items = items
        self.ind = skip

    def skip(self, count: int = 0):
        self.ind += count
        return self

    def __iter__(self):
        return self

    def __next__(self):
        if self.ind >= len(self.items):
            raise StopIteration()
        item = self.items[self.ind]
        self.ind += 1
        return item

    def __len__(self):
        return len(self.items)

    def __getitem__(self, item: int):
        return self.items[item]


class ImageDataStore:
    __slots__ = ["images", "file_name", "ext"]

    def __init__(self, images, file_name, ext):
        self.images = images
        self.file_name = file_name
        self.ext = ext


class Transform:
    def apply(self, image_data: ImageDataStore) -> ImageDataStore:
        raise NotImplementedError()


class ResizeAll(Transform):
    def __init__(self, dims: Tuple[int, int]):
        self.dims = dims

    def apply(self, image_data: "ImageDataStore") -> "ImageDataStore":
        image_data.images = resize_images(image_data.images, self.dims)
        return image_data


class Resize(Transform, Enum):
    UP = 3
    DOWN = 4
    FIRST = 5
    NONE = 0

    def filter(self):
        return {
            Resize.FIRST: return_first,
            Resize.UP: return_max,
            Resize.DOWN: return_min,
        }.get(self, return_first)

    def choose(self, dims: Iterable[Tuple[int, int]]) -> Tuple[int, int]:
        return self.filter()(dims)

    def __str__(self):
        return self.name.lower()

    def apply(self, image_data: ImageDataStore) -> ImageDataStore:
        dims = self.choose(
            [(image.shape[1], image.shape[0]) for image in image_data.images]
        )
        image_data.images = resize_images(image_data.images, dims)
        return image_data


class Stacking(Transform):
    __slots__ = ["cols", "rows", "mode"]

    def __init__(self, cols, rows, mode):
        self.cols = cols
        self.rows = rows
        self.mode = mode

    @classmethod
    def default(cls):
        return cls(1, 1, "rd")

    def apply(self, image_data: ImageDataStore) -> ImageDataStore:
        """
        Expects an array of images a tuple (x,y) that represents how the images will be stacked, and a mode representing
        how the array will be stacked:

        eg. images = [img]*6, dimensions = (2,3), mode='rd':
        2 images per row, 3 rows, ordered from left to right, up to down
        :param image_data:       A set of opened images.
        :return:
        """
        x, y = self.cols, self.rows
        mode = self.mode

        images = image_data.images
        images_stacked = [[None] * x for _ in range(y)]

        # TODO: Excessive quantities of magic occurring here
        if mode[0] in ("l", "r"):
            for i in range(x * y):
                images_stacked[i // x if mode[1] == "d" else y - i // x - 1][
                    i % x if mode[0] == "r" else x - i % x - 1
                ] = images[i]
        elif mode[0] in ("u", "d"):
            for i in range(x * y):
                images_stacked[i % y if mode[0] == "d" else y - i % y - 1][
                    i // y if mode[1] == "r" else x - i // y - 1
                ] = images[i]
        image_data.images = [
            np.concatenate(
                tuple([np.concatenate(tuple(row), axis=1) for row in images_stacked]),
                axis=0,
            )
        ]
        return image_data


class GenericImageIterator:
    def __init__(self, stacking: Stacking = Stacking.default()):
        """
        Samples filtered images from the provided directories in order of appearance.
        Pass in paths to directories containing images, and extension
        of desired image inputs. Will return as many images as needed to fill the stacking.
        May sample unevenly.

        :param directories: Input directories.
        :param exts:        Extension(s) of file to return
        :param num:         Number of images to return each iteration.
        """
        self.transforms: List[Transform] = [stacking]
        self._max_iters = 0
        self._num_iters = 0

    def _name_file(self):
        return str(self._num_iters), None

    def _replace_transform(self, transform: Transform):
        target_ind = find(self.transforms, lambda x: isinstance(x, transform.__class__))
        if target_ind is not None:
            self.transforms[target_ind] = transform
        else:
            self.transforms.append(transform)

    def resize_in(self, dims: Tuple[int, int]):
        if type(self.transforms[0]) not in {Resize, ResizeAll}:
            self.transforms.insert(0, ResizeAll(dims))
        else:
            self.transforms[0] = ResizeAll(dims)
        return self

    def resize_individual(self, resize: Resize):
        if type(self.transforms[0]) not in {Resize, ResizeAll}:
            self.transforms.insert(0, resize)
        else:
            self.transforms[0] = resize
        return self

    def skip(self, items: int):
        raise NotImplementedError()

    def take(self, count: int):
        """
        Sets the number of files that this iterator will output:
        If items <= 0, nothing will be returned. If max_iters >= self.items, self.max_iters remains unchanged.
        Otherwise, self.max_iters = max_iters

        :param count:   The maximum number of iterations that should occur.
        :return:
        """

        self._max_iters = count
        return self

    def chain(self, iterator):
        return ChainIterator([self, iterator])

    def size(self) -> Optional[Tuple[int, int]]:
        stacking_index = find(self.transforms, lambda t: isinstance(t, Stacking))
        stacking = self.transforms[stacking_index]
        index = find(self.transforms, lambda t: isinstance(t, ResizeAll))
        if index is not None:
            dims = self.transforms[index].dims
            return dims[0] * stacking.cols, dims[1] * stacking.rows
        return None

    def choose_padding(self) -> int:
        """
        Chooses the minimum padding to ensure that counter filenames are equal in length.
        :return:
        """
        return len(str(self._max_iters - 1))

    def write_images(
        self,
        dir: str = "./",
        prefix: str = "",
        ext: Optional[str] = None,
        pad: Optional[int] = None,
    ):
        """
        Takes all the images generated by `self` and writes them to their specified file path.
        If no extension is provided, defaults to image file extension, and then to .jpg
        Use self.choose_padding() to ensure that filenames generated through a counter (eg. "0", "1", ...),
        are padded to the same length.

        :param dir:     The target directory
        :param prefix:  A filename prefix
        :param ext:     The desired output extension
        :param pad:     Number of letters to pad the filename with
        :return:
        """
        if ext:
            ext = fo.prepend_dot(ext)

        for image_data in self:
            file_name = image_data.file_name.zfill(pad) if pad else image_data.file_name
            path = (Path(dir) / (prefix + file_name)).with_suffix(
                ext or image_data.ext or ".jpg"
            )
            path.parent.mkdir(exist_ok=True)
            cv2.imwrite(str(path), image_data.images[0])

    def write_video(self, path: str, video_format: str = "mp4v", fps: float = 24.0):
        if self.size() is None:
            raise ValueError(
                f"{self.__class__.__name__} must be initialized with .resize_in() or .resize_out() before calling .write_video()"
            )

        supported_extensions = [".mp4"]
        if Path(path).suffix.lower() not in supported_extensions:
            raise ValueError(
                "Extension %s is not currently supported."
                % (Path(path).suffix.lower(),)
            )

        video_format = cv2.VideoWriter_fourcc(*video_format)
        # apiPreference may be required depending on cv2 version.

        vid = cv2.VideoWriter(
            filename=path,
            apiPreference=0,
            fourcc=video_format,
            fps=fps,
            frameSize=self.size(),
        )

        for image_data in self:
            vid.write(image_data.images[0])
        vid.release()

    def __iter__(self):
        return self

    def __next__(self):
        raise NotImplementedError()

    def _apply_transforms(self, images: ImageDataStore):
        for transform in self.transforms:
            images = transform.apply(images)
        return images


class ChainIterator(GenericImageIterator):
    def __init__(self, iterators: List[GenericImageIterator]):
        super().__init__()
        self.iterators = iterators[::-1]
        self._max_iters = sum(it._max_iters for it in self.iterators)
        self.num_iterators = len(self.iterators)

    def skip(self, items: int):
        while self.iterators and items > 0:
            last = self.iterators[-1]
            if last._max_iters > items:
                items -= last._max_iters
                self.iterators.pop()
            else:
                last.skip(items)
                break

    def chain(self, iterator: GenericImageIterator):
        self.iterators.insert(0, iterator)
        self._max_iters += iterator._max_iters
        self.num_iterators += 1
        return self

    def _name_file(self):
        return str(self._num_iters), None

    def __next__(self):
        while self.iterators:
            try:
                image_data = next(self.iterators[-1])
                image_data = self._apply_transforms(image_data)
                image_data.file_name = (
                    f"{self.num_iterators - len(self.iterators)}_{image_data.file_name}"
                )
                self._num_iters += 1
                return image_data
            except StopIteration:
                self._num_iters -= 1
                self.iterators.pop()
        raise StopIteration


class FileIterator(GenericImageIterator):
    """
    Combines images in order of appearance in each of the provided lists of paths
    """

    def __init__(
        self,
        paths: Union[List[str], List[List[str]]],
        stacking: Stacking = None,
    ):
        if paths and isinstance(paths[0], str):
            paths = [paths]

        super().__init__(stacking=stacking or Stacking(len(paths), 1, "rd"))

        if not all(fo.check_files_exist(paths_) for paths_ in paths):
            raise ValueError("One or more files not found.")

        self.files = [SkipIterator(paths_) for paths_ in paths]
        self._max_iters = max(len(paths_) for paths_ in paths)

        num_images = stacking.cols * stacking.rows
        if num_images != len(paths):
            raise ValueError(
                f"Number of file lists provided ({len(paths)}) does not match number required to fill frame ({num_images} file lists required)"
            )

    def skip(self, count: int):
        for files in self.files:
            files.skip(count)
        return self

    def __next__(self):
        """Returns num images in an array."""
        if self._num_iters >= self._max_iters:
            raise StopIteration()

        output = [cv2.imread(next(files)) for files in self.files]
        name, ext = self._name_file()
        images = ImageDataStore(output, name, ext)
        self._num_iters += 1

        return self._apply_transforms(images)


class DirectoryIterator(FileIterator):
    """
    Combines images in order of appearance in each of the provided directories
    """

    def __init__(
        self, directories, exts=("jpg",), stacking: Stacking = Stacking.default()
    ):
        if isinstance(directories, str):
            directories = [directories]
        if not directories:
            raise ValueError("No directories provided.")

        if not fo.check_dirs_exist(directories):
            raise ValueError("One or more provided directories do not exist.")

        num_images = stacking.cols * stacking.rows
        if num_images != len(directories):
            raise ValueError(
                f"Number of directories provided ({len(directories)}) does not match number required to fill frame ({num_images} directories required)"
            )

        super().__init__(self._load_dirs(directories, exts), stacking)

    def _load_dirs(self, directories: List[str], exts: List[str]):
        files = []
        for directory in directories:
            files.append(fo.get_files(directory, exts))
            if len(files[-1]) == 0:
                raise ValueError(
                    f"No images matching ext {', '.join(exts)} found in {directory}"
                )
        return files


class DirectoryIteratorMatchNames(DirectoryIterator):
    """
    Produces images ordered lexographically by filename, where each image consists of images with matching filenames
    in each of the provided directories.
    """

    def _load_dir(self, directory: str, exts: List[str]):
        files = fo.get_files(directory, exts)

        if not files:
            raise ValueError(f"No files found in {directory}")

        return {os.path.basename(f_name): f_name for f_name in files}

    def _load_dirs(self, directories: List[str], exts: List[str]):
        """Finds and stores all file names which exist across all directories."""
        possible_file_names = set()
        candidates = []
        for i, directory in enumerate(directories):
            files = self._load_dir(directory, exts)
            if i == 0:
                possible_file_names = set(files.keys())
            else:
                possible_file_names &= set(files.keys())

            if not possible_file_names:
                raise ValueError("No file names in common.")

            candidates.append(files)

        common_files = sorted(list(possible_file_names))
        return [
            SkipIterator([candidate[common_file] for common_file in common_files])
            for candidate in candidates
        ]

    def _name_file(self):
        path = Path(self.files[0][self._num_iters])
        return path.stem, path.suffix


def resize_images(images, dimensions: Tuple[int, int]):
    """
    Resizes all of the images to the specified dimensions.
    :param images:          A set of opened images.
    :param dimensions:      The dimensions to resize the images to.
    :return:
    """
    return [cv2.resize(img, dimensions) for img in images]


def make_image_from_images(
    files_in: Union[List[str], str],
    dir_out="./",
    file_name="output",
    ext_out="jpg",
    stacking: Stacking = Stacking.default(),
    size: Tuple[int, int] = (640, 480),
):
    """

    :param files_in:    List of files to read and place into the image.
    :param dir_out:     The directory to output the file(s) to. If it does not exist, it will be created.
    :param file_name:   The initial portion of the filename common to each file.
    :param ext_out:     Output extension for the images.
    :param stacking:    A Stacking object, which defines how the component images should be stacked.
    :param size:        Dimensions of each component image in px.
    :return:
    """
    file_name = fo.form_file_name(dir_out, file_name, ext_out)
    os.makedirs(os.path.dirname(file_name), exist_ok=True)

    if isinstance(files_in, str):
        files_in = [files_in]

    images = [cv2.imread(file) for file in files_in]
    cv2.imwrite(file_name, stacking.apply(resize_images(images, size)))


def dimensions(files: Union[List[str], str]):
    """
    Returns the appropriate dimensions given resize.
    :param files:    One or more files with dimensions of interest
    :return:
    """
    if isinstance(files, str):
        files = [files]

    for file in files:
        ext = Path(file).suffix.lstrip(".")
        ext_cat = fo.FileCategory.from_str(ext)
        if ext_cat == fo.FileCategory.IMAGE:
            yield get_img_dimensions(file)
        else:
            yield get_video_dimensions(file)


def get_dimensions_dirs(
    dirs_in: Union[List[str], str], exts: Union[List[str], str], resize: Resize
):
    """
    Returns the appropriate dimensions given resize.
    :param dirs_in:     One or more directories with files of interest
    :param exts:        The file extension(s).
    :param resize:      A Resize enum.
    :return:
    """
    if isinstance(dirs_in, str):
        dirs_in = [dirs_in]

    def dimension_generator():
        for dir_in in dirs_in:
            try:
                yield resize.choose(dimensions(fo.get_files(dir_in, exts)))
            except ValueError:
                continue

    try:
        return resize.choose(dimension_generator())
    except ValueError:
        if isinstance(exts, str):
            exts = [exts]
        raise ValueError(
            "No files with given extension(s) %s found in any directory."
            % (", ".join(exts),)
        )


def get_img_dimensions(img: str):
    """
    Given a list of file paths, returns the dimensions of the images corresponding to the file paths
    :param img:         File path pointing to image file.
    :return:            List of corresponding file dimensions.
    """
    file = cv2.imread(img)
    return file.shape[1], file.shape[0]


def get_video_dimensions(video: str):
    """
    Given a list of file paths, returns the dimensions of the videos corresponding to the file paths.

    :param video:       File path pointing to video file.
    :return:            List of corresponding file dimensions.
    """
    file = cv2.VideoCapture(video)
    dims = (
        int(file.get(cv2.CAP_PROP_FRAME_WIDTH)),
        int(file.get(cv2.CAP_PROP_FRAME_HEIGHT)),
    )
    file.release()
    return dims


def return_first(items):
    """
    Returns the first element in an array.
    :param items:       Any iterable.
    :return:            The first item in the iterable, or None if the iterable is empty.
    """
    try:
        return next(items)
    except TypeError:
        return next(iter(items))
    except StopIteration:
        return None


def return_max(list_of_dims: Iterable[Tuple[int, int]]) -> Tuple[int, int]:
    """
    Returns the dimensions that produce the maximum area.
    In the event of a tie, will return the first match.

    :param list_of_dims: A list of dimensions.
    :return:             The dimensions with the greatest area.
    """
    return max(list_of_dims, key=lambda dim: dim[0] * dim[1], default=None)


def return_min(list_of_dims: Iterable[Tuple[int, int]]) -> Tuple[int, int]:
    """
    Returns the dimensions that produce the minimum area.
    In the event of a tie, will return the first match.

    :param list_of_dims: A list of dimensions.
    :return:             The dimensions with the minimum area.
    """
    return min(list_of_dims, key=lambda dim: dim[0] * dim[1], default=None)
